/*
 * Copyright (C) Sapphirecode - All Rights Reserved
 * This file is part of snippeteer which is released under BSD-3-Clause.
 * See file 'LICENSE' for full license details.
 * Created by Timo Hocker <timo@scode.ovh>, January 2020
 */

/* eslint-disable no-sync */
/* eslint-disable no-console */

'use strict';

const fs = require ('fs-extra');
const path = require ('path');
const child_process = require ('child_process');

/**
 * copies the full template to a new folder named after arg[0]
 *
 * @param {string} folder folder to run in
 * @param {Array} args function arguments
 */
async function run (folder, args) {
  const snip_folder_path = [ folder ];
  if (args.length > 0)
    snip_folder_path.push (args[0]);
  const snip_folder = path.join (...snip_folder_path);
  const template = path.join (__dirname, 'template');
  if (!fs.existsSync (snip_folder))
    fs.mkdir (snip_folder);
  for (const f of fs.readdirSync (template)) {
    fs.copy (
      path.join (template, f),
      path.join (snip_folder, f),
      {
        recursive: true,
        filter:    (src, dest) => !fs.existsSync (dest)
      }
    );
  }

  child_process.execSync (
    'git init',
    { cwd: snip_folder, stdio: 'inherit' }
  );
  child_process.execSync (
    'yarn init -y',
    { cwd: snip_folder, stdio: 'inherit' }
  );
  child_process.execSync (
    'yarn add --dev @scode/eslint-config eslint nyc ava',
    { cwd: snip_folder, stdio: 'inherit' }
  );

  const package_json = JSON.parse (
    await fs.readFile (path.join (snip_folder, 'package.json'), 'utf-8')
  );

  package_json.scripts = { lint: 'eslint .', test: 'nyc ava' };

  await fs.writeFile (
    path.join (snip_folder, 'package.json'),
    JSON.stringify (package_json, null, 2)
  );

  await fs.writeFile (
    path.join (snip_folder, '.gitignore'),
    `/node_modules/
/dist/`
  );
}

/**
 * checks if the arguments meet the requirements
 *
 * @param {string} folder folder to run in
 * @param {Array} args function arguments
 * @returns {boolean} true if arguments match requirements
 */
function assert (folder, args) {
  const tests = [
    {
      f:      () => (args.length < 2),
      reason: 'too many arguments'
    },
    {
      f:      () => (args.length === 0 || typeof args[0] === 'string'),
      reason: 'name is not a string'
    },
    {
      f:      () => (args.length === 0 || (/^[a-z-]+$/iu).test (args[0])),
      reason: 'name can only contain [a-z-]'
    },
    {
      f:      () => (typeof folder === 'string'),
      reason: 'cwd is not a folder (internal error)'
    },
    {
      f:      () => (fs.existsSync (folder)),
      reason: 'cwd does not exist (internal error)'
    },
    {
      f:      () => (args.length === 1 || fs.readdirSync (folder).length === 0),
      reason: 'folder is not empty'
    },
    {
      f:      () => (args.length === 0
          || !fs.existsSync (path.join (folder, args[0]))),
      reason: 'folder already exists'
    }
  ];
  for (const test of tests) {
    if (!test.f ()) {
      console.log (test.reason);
      return false;
    }
  }

  return true;
}

module.exports = { run, assert };
